
import ldap
from DSGlobals import *
from ldap.filter import filter_format
from DSNodeTypes import *
from copy import deepcopy

class DSContainer(DSNode):
    """Represents a container object in the directory.
    
    Containers may contain nodes or other containers. Adding and deleting 
    of nodes and Containers is done through their parent Container.
    """

    _interfaces = DSNode._interfaces + ("DSContainer",)

    def __init__(self, parent, ldapData):
        """Create a new DSContainer.

        DSContainer objects should only need to be created by DSConnection and
        other DSContainer objects. Create a new DSContainer, passing a
        DSConnection and the ldap dn of the container. Optionally, the ou name
        may be provided, it will be used as the display name. A description may
        also be provided, and will be stored as container.description.
        """

        self._dsconn = parent._dsconn
        self._parent = parent
        self.conn = parent.conn

        self._dn = ldapData[0]
        self._objectClasses = ldapData[1].get('objectClass', ['organizationalUnit'])
        self.displayName = ldapData[1].get('displayName', [None])[0]
        self._description = ldapData[1].get('description', [None])[0]
        self._ou = ldapData[1].get('ou', [None])[0]

        self._cached_attributes = None
        self._new = None
        
        self._child_node_cache = None
        self._child_container_cache = None



    # Make description a property so we can document it
    description = property((lambda self: self._description),
                            doc="Description of the container")
    # Make dsconn a property so we can document it
    dsconn = property((lambda self: self._dsconn),
                            doc="A handle to the DSConnection object")

    def Parent(self):
        if self._dn == self.dsconn.basedn:
            return self
        path = ldap.explode_dn(self._dn)
        parent_dn = ','.join(path[1:])
        return DSContainer(self, (parent_dn, {}))

    parent = property((lambda self: self.Parent()),
                        doc="The parent container of this container")

    def __repr__(self):
        return '<DS: %s>' % (self.Name())

    def Name(self):
        """Return a name for the container.

        Returns a name suitable for display to a user if one was provided (as
        the ou argument to the constuctor), otherwise returns the dn of the
        container.
        """
        if (self._ou):
            return self._ou
        else:
            return self._dn

    def Path(self):
        """Return the ldap path (dn) of the container."""
        return self._dn

    def getChildNodes(self):
        """Return the non-container children of this container as a list of DSNode objects."""
        
        if (not self._child_node_cache):
            self.Refresh()

        return self._child_node_cache

    def getChildContainers(self):
        """Return a list of the container objects within the container."""
        
        if (not self._child_container_cache):
            self.Refresh()
        
        return self._child_container_cache
        
    def Refresh(self):
        """Performs a new ldap query to update the cached child information"""

        # We will search for all types of objects, and split them into nodes and
        # containers based on the returned attributes
        resid = self.conn.search(self._dn, ldap.SCOPE_ONELEVEL, \
            '(objectClass=*)', ('ou','objectClass','displayName','uid', 'description'))
        
        try:
            rescode, results = self.conn.result(resid)
        except ldap.INSUFFICIENT_ACCESS, err:
            raise InsufficientAccessError()
        except Exception, err:
            raise DSError("Error listing child containers of '%s': '%s'" % \
                    (self._dn, err))

        containers = []
        nodes = []
        for child in results:
            child_dn = child[0]
            if('organizationalUnit' in child[1]['objectClass']):
                # We put the container objects in the containers list
                containers.append(DSContainer(self, child))
            else:
                # and the rest in the nodes list
                nodes.append(NodeFactory(self, child))
        
        self._child_container_cache = containers
        self._child_node_cache = nodes

    def CreateUser(self, uid, primaryGroup, givenName=None, sn=None, 
                    initials='', cn=None, displayName=None, 
                    description='', homeDirectory=None, gecos='', 
                    uidNumber=None):
        """Create a new user account node within this container.

        A uid (string) and primary group (DSNode) must be provided. posixGroup
        and inetOrgPerson objectclasses will be used, and if possible,
        sambaSAMAccount objectclass.

        If no uidNumber is specified, it will search for a UnixIDPool object
        and use the next gidNumber specified by that object. If no ID Pool is
        found, it will search for all users in the directory, find the highest
        existing uidNumber, and increment it.
        """

        userdn = "uid=%s,%s" % (uid, self._dn)
        gidNumber = primaryGroup._single_get("gidNumber")


        # Check whether an an account with this name exists
        filter = filter_format("(&(objectClass=posixAccount)(uid=%s))", [uid])
        nodes = self.dsconn._search(filter)
        if (nodes):
            raise DSError("A user '%s' already exists." % (uid))

        if not (cn):
            cn = uid

        if not (displayName):
            displayName = cn

        if not (givenName):
            givenName = uid
            
        if not (homeDirectory):
            homeDirectory = '/home/%s' % (uid)

        if not (sn):
            sn = 'n/a'

        IDPool = None
        if (not uidNumber):
            # Search for an ID Pool object and get the uidNumber from it
            filter = IDSUnixIDPool.ldapSearchFilter
            res = self.dsconn._search(filter)
            if (res):
                IDPool = res[0]
                uidNumber = IDPool['UidNumber']
                IDPool['UidNumber'] = str(int(uidNumber) + 1)
                # SetInfo is run after the account is created

            else:
                # There was no "NextFreeUnixId" object to use, do it the hard
                # way: get every single user account and ++ the largest
                # uidNumber yes, it is ugly...
                resid = self.conn.search(self.dsconn.basedn, \
                    ldap.SCOPE_SUBTREE, '(objectClass=posixAccount)', \
                    ['uidNumber'])

                rescode, results = self.conn.result(resid)

                uidNumbers = [o[1].get('uidNumber', ['0'])[0] for o in results]
                uidNumbers.sort()
                uidNumber = str( int(uidNumbers[-1]) + 1 )

        assert(uidNumber)

        newAttributes = {'objectClass': ['posixAccount', 'inetOrgPerson'],
                    'uid' : uid,
                    'cn' : cn,
                    'uidNumber' : uidNumber,
                    'gidNumber' : gidNumber,
                    'givenName' : givenName,
                    'sn' : sn,
                    'initials' : initials,
                    'description' : description,
                    'displayName' : displayName,
                    'homeDirectory' : homeDirectory,
                    'gecos' : gecos }
                    
        modlist = ldap.modlist.addModlist(newAttributes)
        resid = self.conn.add(userdn, modlist)
        rescode, results = self.conn.result(resid)
        
        if (IDPool):
            IDPool.SetInfo()

        # Try to add samba objectclass
        searchstring = "(objectClass=sambaDomain)"
        nodes = self.dsconn._search(searchstring)
        if (nodes):
            sambaDomain=nodes[0]
            domainSID = sambaDomain._single_get("sambaSID")
            sambaAttributes = deepcopy(newAttributes)
            sambaAttributes['objectClass'].append('sambaSAMAccount')
            sambaSID = domainSID + "-" + str(int(uidNumber) * 2 + 1000)
            sambaAttributes['sambaSID'] = sambaSID
            sambaAttributes['sambaAcctFlags'] = '[UX]'
            modlist = ldap.modlist.modifyModlist(newAttributes, sambaAttributes)
            try:
                self.conn.modify_s(userdn, modlist)
                newAttributes = sambaAttributes
            except:
                pass

        self.Refresh()

    def CreateUserList(self, name, memberNodes):
        """Create a userlist object with the specified nodes as members."""

        userlistdn = "cn=%s,%s" % (name.replace(',', '\\,'), self._dn)
        members = [m._dn for m in memberNodes]

        newAttributes = {'objectClass': ['groupOfNames'],
                    'cn' : name,
                    'member'    : members}

        modlist = ldap.modlist.addModlist(newAttributes)
        resid = self.conn.add(userlistdn, modlist)
        rescode, results = self.conn.result(resid)
        self.Refresh()
        

    def CreateGroup(self, name, gidNumber=None):
        """Create a group object.
        
        If no gidNumber is specified, it will search for a UnixIDPool object
        and use the next gidNumber specified by that object. If no ID Pool is
        found, it will search for all groups in the directory, find the highest
        existing gidNumber, and increment it.
        """

        filter = filter_format("(&(objectClass=posixGroup)(cn=%s))", [name])
        nodes = self.dsconn._search(filter)
        if (nodes):
            raise DSError("A group '%s' already exists." % (name))

        groupdn = "cn=%s,%s" % (name.replace(',', '\\,') , self._dn)
        IDPool = None

        if (not gidNumber):
            filter = IDSUnixIDPool.ldapSearchFilter
            res = self.dsconn._search(filter)
            if (res):
                IDPool = res[0]
                print IDPool.PrettyPath()
                print IDPool._objectClasses
                print IDPool.Implements()
                gidNumber = IDPool['GidNumber']
                IDPool['GidNumber'] = str(int(gidNumber) + 1)
                # SetInfo is run after the account is created

            else:
                # There was no "NextFreeUnixId" object to use, do it the
                # hard way: get every single user account and ++ the
                # largest gidNumber yes, it is ugly...
                
                # We use this kind of search because otherwise there
                # would be way too many calls to the ldap server
                # as we GetInfo every posixGroup object
                resid = self.conn.search(self.dsconn.basedn, \
                    ldap.SCOPE_SUBTREE, '(objectClass=posixGroup)', \
                    ['gidNumber'])
                rescode, results = self.conn.result(resid)

                gidNumbers = [o[1].get('gidNumber', ['0'])[0] for o in results]
                gidNumbers.sort()
                gidNumber = str( int(gidNumbers[-1]) + 1 )

        newAttributes = {'objectClass': ['posixGroup'],
                    'cn' : name,
                    'gidNumber' : gidNumber,
        }
                    
        modlist = ldap.modlist.addModlist(newAttributes)
        resid = self.conn.add(groupdn, modlist)
        rescode, results = self.conn.result(resid)

        if (IDPool):
            IDPool.SetInfo()
        
        # Try to add samba objectclass
        filter = IDSSambaDomain.ldapSearchFilter
        res = self.dsconn._search(filter)
        if (res):
            sambaDomain = res[0]
            domainSID = sambaDomain["SambaSID"]
            sambaAttributes = deepcopy(newAttributes)
            sambaAttributes['objectClass'].append('sambaGroupMapping')
            sambaSID = domainSID + "-" + str(int(gidNumber) * 2 + 1001)
            sambaAttributes['sambaSID'] = sambaSID
            sambaAttributes['sambaGroupType'] = '2'
            modlist = ldap.modlist.modifyModlist(newAttributes, sambaAttributes)
            try:
                self.conn.modify_s(groupdn, modlist)
                newAttributes = sambaAttributes
            except:
                pass

        self.Refresh()

    def CreateFolder(self, name):
        folderdn = "ou=%s,%s" % (name.replace(',', '\\,'), self._dn)

        newAttributes = {'objectClass': ['organizationalUnit'],
                    'ou' : name}

        modlist = ldap.modlist.addModlist(newAttributes)
        resid = self.conn.add(folderdn, modlist)
        rescode, results = self.conn.result(resid)
        self.Refresh()

    def Delete(self):
        for child in self._child_container_cache:
            child.Delete()
        for child in self._child_node_cache:
            if child.Implements('IDSUser'):
                child.Delete()
        for child in self._child_node_cache:
            if not child.Implements('IDSUser'):
                child.Delete()
        DSNode.Delete(self)

    def DeleteNode(self, node):
        """Delete the given node.

        Attempts to maintain referential integrity for the known types by e.g.
        removing users from the groups they are members of before deleting,
        catching the fact that a group is being used as someone's primary
        group. etc.
        """
        return node.Delete()

def DSContainerTest(parent, dn, objectClasses):
    if 'organizationalUnit' in objectClasses:
        return DSContainer
NodeTypeTests.append(DSContainerTest)
